import os

curr_path = os.path.dirname(os.path.abspath(__file__))

TOKEN_FILE_PATH = os.path.expanduser("~/.hub/token")
CACHE_FILE_PATH = os.path.expanduser("~/.hub/tmp")
CLOUDVOLUME_PATH = os.path.expanduser("~/.cloudvolume/secrets/aws-secret.json")
AWSCRED_PATH = os.path.expanduser("~/.aws/credentials")



import hub


def test_dynamic_array():
    conn = hub.fs("./data/cache").connect()
    arr = conn.array_create(
        "test/dynamic_array_3",
        shape=(10, 8, 4, 12),
        chunk=(5, 4, 2, 6),
        dtype="uint8",
        dsplit=2,
    )
    arr.darray[0:10, 0:5] = (2, 12)
    arr.darray[0:10, 5:8] = (6, 14)

    assert arr[5, 3, :, :].shape == (2, 12)
    assert arr[5, 6, :, :].shape == (6, 14)
    assert arr[5, 4:6, :, :].shape == (2, 6, 14)



from typing import *
import os, sys, time, random, json, itertools, uuid, traceback

from .bucket import Bucket
from . import storage
from .storage.retry_wrapper import RetryWrapper
from hub.utils.store_control import StoreControlClient
import configparser
from hub.exceptions import S3CredsParseException


class Base:
    def __init__(self):
        pass

    def connect(self) -> Bucket:
        return Bucket(RetryWrapper(self._create_storage()))

    @staticmethod
    def _s3(
        bucket: Optional[str] = None,
        aws_creds_filepath: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
    ) -> "creds.Base":
        if aws_access_key_id is not None or aws_secret_access_key is not None:
            assert aws_access_key_id is not None
            assert aws_secret_access_key is not None
            assert bucket is not None
            assert aws_creds_filepath is None
            return S3(bucket, aws_access_key_id, aws_secret_access_key)
        elif aws_creds_filepath is not None:
            config = configparser.ConfigParser()
            path = os.path.expanduser(aws_creds_filepath)
            path = os.path.abspath(path)
            config.read(path)

            if len(config.sections()) > 0:
                default = config.sections()[0]
                aws_access_key_id = config[default]["aws_access_key_id"]
                aws_secret_access_key = config[default]["aws_secret_access_key"]
            else:
                with open(aws_creds_filepath, "r") as f:
                    j = json.loads(f.read())
                    if not bucket:
                        bucket = j["bucket"]
                    aws_access_key_id = j["aws_access_key_id"]
                    aws_secret_access_key = j["aws_secret_access_key"]
            return S3(bucket, aws_access_key_id, aws_secret_access_key)
        else:
            # Try load from IAM role
            try:
                return S3(bucket, aws_access_key_id, aws_secret_access_key)
            except Exception as e:
                print("using default creds")
            config = StoreControlClient.get_config()
            if bucket is None:
                bucket = config["BUCKET"]

            aws_access_key_id = config["AWS_ACCESS_KEY_ID"]
            aws_secret_access_key = config["AWS_SECRET_ACCESS_KEY"]

            return S3(bucket, aws_access_key_id, aws_secret_access_key)

    def s3(
        self,
        bucket: Optional[str] = None,
        aws_creds_filepath: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
    ) -> "creds.Base":
        return Recursive(
            self,
            self.__class__._s3(
                bucket, aws_creds_filepath, aws_access_key_id, aws_secret_access_key
            ),
        )

    @staticmethod
    def _gs(bucket: str, creds_path: Optional[str] = None):
        return GS(bucket, creds_path)

    def gs(self, bucket: str, creds_path: Optional[str] = None) -> "creds.Base":
        return Recursive(self, self.__class__._gs(bucket, creds_path))

    @staticmethod
    def _fs(dir: str) -> "creds.Base":
        dir = os.path.expanduser(dir)

        if not os.path.exists(dir):
            os.makedirs(dir)

        return FS(dir)

    def fs(self, dir: str) -> "creds.Base":
        return Recursive(self, self.__class__._fs(dir))

    def _create_storage(self) -> storage.Base:
        raise NotImplementedError()


class Recursive(Base):
    _core: "creds.Base" = None
    _wrapper: "creds.Base" = None

    def __init__(self, core: "creds.Base", wrapper: "creds.Base"):
        super().__init__()
        self._core = core
        self._wrapper = wrapper

    def _create_storage(self):
        return storage.Recursive(
            self._core._create_storage(), self._wrapper._create_storage()
        )


class S3(Base):

    _bucket: str = None
    _aws_access_key_id: str = None
    _aws_secret_access_key: str = None

    def __init__(self, bucket: str, aws_access_key_id: str, aws_secret_access_key: str):
        super().__init__()
        self._bucket = bucket
        self._aws_access_key_id = aws_access_key_id
        self._aws_secret_access_key = aws_secret_access_key

    def _create_storage(self):
        return storage.S3(
            self._bucket, self._aws_access_key_id, self._aws_secret_access_key
        )


class GS(Base):
    _creds_path: str = None
    _bucket: str = None

    def __init__(self, bucket: str, creds_path: str):
        self._bucket = bucket
        if creds_path is not None:
            self._creds_path = os.path.expanduser(creds_path)
        else:
            self._creds_path = None

    def _create_storage(self):
        return storage.GS(self._bucket, self._creds_path)


class FS(Base):
    _dir: str = None

    def __init__(self, dir: str):
        super().__init__()
        self._dir = dir

    def _create_storage(self):
        return storage.FS(self._dir)



import os
import shutil

from hub.config import CACHE_FILE_PATH
from hub.exceptions import FileSystemException

from .base import Base

class FS(Base):
    def __init__(self, dir: str):
        super().__init__()
        self._dir = dir

    def get(self, path: str) -> bytes:
        path = os.path.join(self._dir, path)        
        with open(path, 'rb') as f:
            return f.read()

    def put(self, path: str, content: bytes):
        path = os.path.join(self._dir, path)
        folder = '/'.join(path.split('/')[:-1])
        if not os.path.isdir(folder):
            os.makedirs(folder)

        with open(path, 'wb') as f:
            f.write(content)

    def exists(self, path: str) -> bool:
        path = os.path.join(self._dir, path)
        return os.path.exists(path)

    def delete(self, path: str):
        path = os.path.join(self._dir, path)
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)


import hub
import torch
import torch.utils.data as data
import torchvision.transforms as transforms
import numpy as np


class TorchDataset(data.Dataset):
    """ Dataset
    Args:
        dataset (hub.Dataset): hub dataset object 
    """

    def __init__(self, dataset, transforms=None):
        self.size = dataset.shape[0]
        self.key = dataset.key
        self.dataset = None
        self.batch = None
        self.indexbeta = 0
        self.batch_size = dataset.chunk_shape[0]

    def __getitem__(self, index):
        """
        Args:
            index (int): Index
        Returns:
            tuple: (dataset,...) 
        """
        # Load dataset for each worker
        if not self.dataset:
            self.dataset = hub.Dataset(key=self.key)

        # if self.batch is None or self.indexbeta == self.batch_size:
        #     self.indexbeta = 0
        #     self.batch = self.dataset[:self.batch_size]

        # output = list(map(lambda x:x[self.indexbeta], self.batch))
        # self.indexbeta += 1

        output = list(self.dataset[index])

        return (*output,)

    def __len__(self):
        return self.size


class TorchIterableDataset(data.IterableDataset):
    def __init__(self, dataset, transform=None):
        self.size = dataset.shape[0]
        self.path = dataset._path
        self.storage = dataset._storage
        self.dataset = None
        self.transform = transform
        self._components = dataset._components
        self._dynkeys = {
            key for key, value in self._components.items() if value.darray is not None
        }

    def _get_common_chunk(self, dataset):
        batch_sizes = map(lambda x: x[0], dataset.chunks.values())
        return min(batch_sizes)

    def _enumerate(self, dataset):
        """
        Seperate for each worker the id such that there is no conflict
        """
        worker_info = torch.utils.data.get_worker_info()

        if worker_info is None:
            for i, x in enumerate(dataset):
                yield x

        if worker_info is None:
            id = 1
            workers = 1
        else:
            id = worker_info.id
            workers = worker_info.num_workers
        batch = self._get_common_chunk(dataset)

        # Be careful might contain bugs
        for i in range(dataset.shape[0]):
            if (i // batch) % workers == id:
                yield dataset[i]

    def __iter__(self):

        if self.dataset is None:
            self.dataset = hub.Dataset(self.path, self.storage)

        # keys = tuple(self.dataset.paths.keys())
        # dyn_keys = [key for key in keys if self.dataset[key].darray is not None]

        for x in self._enumerate(self.dataset):
            if self.transform:
                x = self.transform(x)
            yield x

    def __len__(self):
        return self.size

    def collate_fn(self, batch):
        batch = tuple(batch)
        keys = tuple(batch[0].keys())
        ans = {key: [item[key] for item in batch] for key in keys}
        for key in keys:
            if key not in self._dynkeys:
                ans[key] = torch.tensor(ans[key])
            else:
                ans[key] = [torch.tensor(item) for item in ans[key]]
        return ans



import json
from typing import Tuple, Dict
from .storage import Base as Storage
from .array import Array, Props


class DatasetProps:
    paths: Dict[str, str] = None


class Dataset:
    def __init__(self, path: str, storage: Storage):
        self._path = path
        self._storage = storage
        self._props = DatasetProps()
        self._props.__dict__ = json.loads(storage.get(path + "/info.json"))
        self._components = self._setup(self.paths)

    @property
    def paths(self) -> Dict[str, str]:
        return self._props.paths

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._common_shape(self.shapes)

    @property
    def chunk(self) -> Tuple[int, ...]:
        return self._common_shape(self.chunks)

    @property
    def shapes(self) -> Dict[str, Tuple[int, ...]]:
        return self._get_property("shape")

    @property
    def chunks(self) -> Dict[str, Tuple[int, ...]]:
        return self._get_property("chunk")

    @property
    def dtype(self) -> Dict[str, str]:
        return self._get_property("dtype")

    @property
    def compress(self) -> Dict[str, str]:
        return self._get_property("compress")

    @property
    def compresslevel(self) -> Dict[str, float]:
        return self._get_property("compresslevel")

    def _get_property(self, name: str) -> Dict:
        return {k: getattr(comp, name) for k, comp in self._components.items()}

    def _common_shape(self, shapes: Tuple[int, ...]) -> Tuple[int, ...]:
        shapes = [shapes[k] for k in shapes]
        shapes = sorted(shapes, key=lambda x: len(x))
        min_shape = shapes[0]
        common_shape = []
        for dim in range(len(min_shape)):
            for shp in shapes:
                if min_shape[dim] != shp[dim]:
                    return common_shape
            common_shape.append(min_shape[dim])
        return common_shape

    def _setup(self, components: Dict[str, str]) -> Dict[str, Array]:
        datas = {}
        if components is None:
            return datas

        for key, path in components.items():
            datas[key] = Array(path, self._storage)

        return datas

    def to_pytorch(self, transform=None):
        try:
            from hub.integrations.pytorch import TorchIterableDataset

            return TorchIterableDataset(self, transform=transform)
        except Exception as ex:
            raise ex  # Exception('PyTorch is not installed')

    def to_tensorflow(self):
        try:
            from hub.integrations.tensorflow import HubTensorflowDataset

            return HubTensorflowDataset(self)
        except Exception as ex:
            raise ex  # Exception('TensorFlow is not intalled')

    def __getitem__(self, slices):
        if not isinstance(slices, list) and not isinstance(slices, tuple):
            slices = [slices]

        if isinstance(slices[0], str):
            if len(slices) == 1:
                return self._components[slices[0]]
            else:
                return self._components[slices[0]][slices[1:]]
        else:
            if len(slices) <= len(self.shape):
                datas = {key: value[slices] for key, value in self._components.items()}
                # return list(map(lambda x: x[slices], datas))
                return datas
            else:
                raise Exception(
                    "Slices ({}) could not much to multiple arrays".format(slices)
                )

    def __setitem__(self, slices, item):
        if isinstance(slices[0], str):
            if len(slices) == 1:
                return self._components[slices[0]]
            else:
                self._components[slices[0]][slices[1:]] = item
        else:
            if len(slices) < len(self.chunk_shape) and len(item) == len(
                self._components
            ):
                datas = [self._components[k] for k in self._components]

                def assign(xy):
                    xy[0][slices] = xy[1]

            return list(map(assign, zip(datas, item)))

    def __len__(self):
        return self.shape[0]

    def items(self):
        return list(self._components.items())

    def __iter__(self):
        for i in range(0, len(self)):
            yield self[i]



import hub
import pytest
import os

# Scenario 1. User has AWS credentials, no hub creds


def test_s3_right_creds():
    try:
        bucket = hub.s3("snark-test", aws_creds_filepath=".secrets/s3").connect()
        bucket.blob_set("temporary_blob.txt", bytes(1))
        bucket.blob_get("temporary_blob.txt")
    except:
        pytest.fail("Unexpected error.. could not connect to s3")


def test_s3_wrong_creds():
    with open("./data/cache/wrong_creds.txt", "w") as f:
        f.write("empty")
    with pytest.raises(Exception):
        bucket = hub.s3(
            "snark-test", aws_creds_filepath="./data/cache/wrong_creds.txt"
        ).connect()
        bucket.blob_set("temporary_blob.txt", bytes(1))
        bucket.blob_get("temporary_blob.txt")



import hub
import pytest


def test_gs():
    bucket = hub.gs("snark-test", creds_path=".secrets/gcs.json").connect()
    bucket.blob_set("temporary_blob.txt", bytes(1))
    text = bucket.blob_get("temporary_blob.txt")
    assert text == bytes(1)


def test_s3_blob():
    bucket = hub.s3("snark-test", aws_creds_filepath=".secrets/s3").connect()
    bucket.blob_set("temporary_blob.txt", bytes(1))
    text = bucket.blob_get("temporary_blob.txt")
    assert text == bytes(1)


def test_s3_array():
    bucket = hub.s3("snark-test", aws_creds_filepath=".secrets/s3").connect()
    x = bucket.array(
        name="test_array", shape=(10, 10, 10), chunk=(5, 5, 5), dtype="uint8"
    )
    x[:] = 2
    arr = bucket.open(name="test_array")
    assert arr[1, 1, 1] == 2


def test_s3_delete_item():
    bucket = hub.s3("snark-test", aws_creds_filepath=".secrets/s3").connect()
    bucket.delete(name="test_array")
    with pytest.raises(Exception):
        bucket.open(name="test/example:5")



import hub
import torch
from torchvision import transforms
import torchvision.datasets as datasets

import torch
import numpy as np
from torchvision import transforms
import random

# from skimage.transform import rotate


def test_pytorch():
    # Create arrays
    datahub = hub.fs("./data/cache").connect()
    images = datahub.array(
        name="test/dataloaders/images3",
        shape=(100, 100, 100),
        chunk=(1, 100, 100),
        dtype="uint8",
    )
    labels = datahub.array(
        name="test/dataloaders/labels3", shape=(100, 1), chunk=(100, 1), dtype="uint8"
    )
    # Create dataset
    ds = datahub.dataset(
        name="test/loaders/dataset2", components={"images": images, "labels": labels}
    )
    # Transform to Pytorch
    train_dataset = ds.to_pytorch()
    # Create data loader
    train_loader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=32,
        num_workers=2,
        pin_memory=False,
        shuffle=False,
        drop_last=False,
    )
    # Loop over attributes
    batch = next(iter(train_loader))
    assert batch["images"].shape == (32, 100, 100)
    assert batch["labels"].shape == (32, 1)


# TODO keras
# def test_keras():
#     print("not implemented")


def test_to_tensorflow():
    print("testing Tensorflow")
    conn = hub.fs("./data/cache").connect()
    ds = conn.open("test/loaders/dataset2")

    # Transform to Tensorflow
    train_dataset = ds.to_tensorflow()
    batch = next(iter(train_dataset.batch(batch_size=16)))
    assert batch["images"].shape == (16, 100, 100)
    # TODO create dataloader



import boto3
import botocore

from .base import Base


class S3(Base):
    def __init__(self, bucket: str, aws_access_key_id: str, aws_secret_access_key: str):

        super().__init__()
        self._bucket_name = bucket
        if aws_access_key_id is None and aws_secret_access_key is None:
            self._client = boto3.client(
                "s3", config=botocore.config.Config(max_pool_connections=128),
            )
            self._resource = boto3.resource("s3")
        else:
            self._client = boto3.client(
                "s3",
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                config=botocore.config.Config(max_pool_connections=128),
            )
            self._resource = boto3.resource(
                "s3",
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
            )
        self._bucket = self._resource.Bucket(bucket)

    def get(self, path: str) -> bytes:
        return self._bucket.Object(path).get()["Body"].read()

    def put(self, path: str, content: bytes) -> None:
        attrs = {
            "Bucket": self._bucket_name,
            "Body": content,
            "Key": path,
            "ContentType": ("application/octet-stream"),
        }
        self._client.put_object(**attrs)

    def exists(self, path: str) -> bool:
        return self.get_or_none(path) is not None

    def delete(self, path: str):
        self._bucket.objects.filter(Prefix=path).delete()

    def get_or_none(self, path):
        try:
            resp = self._client.get_object(Bucket=self._bucket_name, Key=path,)
            return resp["Body"].read()
        except botocore.exceptions.ClientError as err:
            if err.response["Error"]["Code"] == "NoSuchKey":
                return None
            else:
                raise err



import pytest
import hub
import numpy as np


def test_init():
    print("- Initialize array")
    shape = (10, 10, 10, 10)
    chunk = (5, 5, 5, 5)
    datahub = hub.fs("./data/cache").connect()
    x = datahub.array(name="test/example:1", shape=shape, chunk=chunk, dtype="uint8")
    shape = np.array(shape)
    assert np.all(np.array(x.shape) == shape)
    print("passed")


def test_simple_upload_download():
    print("- Simple Chunk Upload and Download")
    datahub = hub.fs("./data/cache").connect()
    shape = (10, 10, 10, 10)
    chunk = (5, 5, 5, 5)
    datahub = hub.fs("./data/cache").connect()
    x = datahub.array(name="test/example:1", shape=shape, chunk=chunk, dtype="uint8")
    x[0] = np.ones((1, 10, 10, 10), dtype="uint8")
    assert x[0].mean() == 1
    print("passed")


def test_multiple_upload_download():
    datahub = hub.fs("./data/cache").connect()
    shape = (10, 10, 10, 10)
    chunk = (5, 5, 5, 5)
    x = datahub.array(name="test/example:1", shape=shape, chunk=chunk, dtype="uint8")
    x[0:3] = np.ones((3, 10, 10, 10), dtype="uint8")
    assert x[0:3].mean() == 1
    print("passed")


def test_cross_chunk_upload_download():
    print("- Cross Chunk Upload and Download")
    datahub = hub.fs("./data/cache").connect()
    shape = (10, 10, 10, 10)
    chunk = (5, 5, 5, 5)
    x = datahub.array(name="test/example:2", shape=shape, chunk=chunk, dtype="uint8")
    x[2:5, 0:9, 0:10] = np.ones((3, 9, 10, 10), dtype="uint8")
    x[2:5, 9:10] = np.zeros((3, 1, 10, 10), dtype="uint8")
    assert x[2:5, 0:9, 0:10].mean() == 1
    assert x[2:5, 9:10].mean() == 0
    print("passed")


def test_broadcasting():
    print("- Broadcasting")
    datahub = hub.fs("./data/cache").connect()
    shape = (100, 100, 100)
    chunk = (50, 50, 50)
    x = datahub.array(name="test/example:3", shape=shape, chunk=chunk, dtype="uint8")
    x[0, 0, 0] = 11
    assert x[0, 0, 0] == 11
    x[0] = 10
    assert x[0].mean() == 10
    x[1] = np.ones((100, 100), dtype="uint8")
    assert x[1].mean() == 1
    x[3, 90, :] = np.ones((1, 1, 100), dtype="uint8")
    assert x[3, 90].mean() == 1
    print("passed")


def test_chunk_shape():
    print("- Chunk shape")
    datahub = hub.fs("./data/cache").connect()
    shape = (10, 10, 10)
    chunk = (5, 5, 5)
    x = datahub.array(name="test/example:4", shape=shape, chunk=chunk, dtype="uint8")
    x[0:5, 0:5, 0:5] = 0
    print("passed")


def test_open_array():
    print("- Loading arrays")
    datahub = hub.fs("./data/cache").connect()
    x = datahub.open(name="test/example:4")
    print(x.shape)
    assert np.all(x.shape == np.array((10, 10, 10)))
    print("passed")


def test_squeeze_array():
    print("- Squeezing arrays")
    datahub = hub.fs("./data/cache").connect()
    shape = (10, 10, 10)
    chunk = (5, 5, 5)
    x = datahub.array(name="test/example:4", shape=shape, chunk=chunk, dtype="uint8")
    assert len(x[0].shape) == 2
    assert len(x[:1].shape) == 3
    assert len(x[0, 0, :].shape) == 1
    assert x[0, 0, 0] == 0
    print("passed")


def test_dtypes():
    print("- Numpy dtypes arrays")
    datahub = hub.fs("./data/cache").connect()
    shape = (10, 10, 10)
    chunk = (5, 5, 5)
    x = datahub.array(name="test/example:5", shape=shape, chunk=chunk, dtype="float32")
    assert x.dtype == "<f4"
    print("passed")


def test_delete_item():
    datahub = hub.fs("./data/cache").connect()
    datahub.delete(name="test/example:5")
    with pytest.raises(Exception):
        datahub.open(name="test/example:5")




from .base import Base

class Recursive(Base):
    _core: Base = None
    _wrapper: Base = None

    def __init__(self, core: Base, wrapper: Base):
        super().__init__()
        self._core = core
        self._wrapper = wrapper

    def get(self, path: str) -> bytes:
        ans = self._wrapper.get_or_none(path) 
        if ans is None:
            ans = self._core.get(path)
            self._wrapper.put(path, ans)
            return ans
        else:
            return ans

    def put(self, path: str, content: bytes):
        self._core.put(path, content)
        self._wrapper.put(path, content)

    def exists(self, path: str) -> bool:
        return self._wrapper.exists(path) or self._core.exists(path)
    
    def delete(self, path: str):
        self._wrapper.delete(path)
        self._core.delete(path)



class OutOfBoundsError(Exception):
    """Raised upon finding a missing chunk."""
    pass


class AlignmentError(Exception):
    """Raised when there is an Alignment error."""
    pass


class IncompatibleShapes(Exception):
    """Shapes do not match"""
    pass


class IncompatibleBroadcasting(Exception):
    """Broadcasting issue"""
    pass


class IncompatibleTypes(Exception):
    """Types can not cast"""
    pass


class WrongTypeError(Exception):
    """Types is not supported"""
    pass


class NotAuthorized(Exception):
    """Types is not supported"""
    pass


class NotFound(Exception):
    """When Info could not be found for array"""
    pass


class FileSystemException(Exception):
    """Error working with local file system"""
    pass


class S3Exception(Exception):
    """Error working with AWS"""
    pass


class S3CredsParseException(Exception):
    """Can't parse AWS creds"""
    pass


from typing import *
import json, os, sys, time, random, uuid, io

from .array import Array, Props
from .storage import Base as Storage
from .dataset import Dataset, DatasetProps

try:
    import torch
except ImportError:
    pass

import numpy as np


class Bucket:
    _storage: Storage = None

    def __init__(self, storage: Storage):
        self._storage = storage

    def array(
        self,
        name: str,
        shape: Iterable[int],
        chunk: Iterable[int],
        dtype: Union[str, np.dtype],
        compress: str = "default",
        compresslevel: float = 0.5,
        dsplit: int = None,
    ) -> Array:
        """Create Hub array.

        Parameters
        ----------
        name: str
            Path to store the array in local storage or in cloud
        shape: tuple
            Shape of the array like (1000, 1000, 100)
        chunk: tuple
            How to chunk the array. Must have the same length as shape. like(100, 100, 10)
        dtype: str or np.dtype
            The desired data type for the array
        Returns
        -------
        Hub Array: Chunkified array stored in local storage or in cloud
        Examples
        --------
        >>> datahub = hub.fs('./data/cache').connect()
        >>> arr = datahub.array('your_array_name', shape=(100000, 512, 512, 3), chunk=(100, 512, 512, 3), dtype='float32')
        >>> arr[0] = np.random.rand((512, 512, 3), dtype='float32')
        """

        return self.array_create(
            name, shape, chunk, dtype, compress, compresslevel, dsplit
        )

    def dataset(self, name: str, components: Dict[str, str]):
        """Create Hub Dataset.

        Parameters
        ----------
        name: str
            Path to store the dataset in local storage or in cloud
        components: dict
            Dictionary of Hub Arrays
        Returns
        -------
        Hub Dataset: Dataset containing Hub Arrays stored in local storage or in cloud
        Examples
        --------
        >>> datahub = hub.fs('./data/cache').connect()
        >>> images = datahub.array('images', shape=(100000, 512, 512, 3), chunk=(100, 512, 512, 3), dtype='float32')
        >>> images[0] = np.random.random((512, 512, 3))
        >>> labels = datahub.array('labels', shape=(100000, 2), chunk=(100, 2), dtype='uint8')
        >>> labels[0] = 1
        >>> ds = datahub.dataset(name="test/dataset:train", components={"images": images, "labels": labels})
        """
        return self.dataset_create(name, components)

    def open(self, name: str):
        jsontext = self._storage.get_or_none(os.path.join(name, "info.json"))
        if jsontext is None:
            return self.blob_get(name)
        else:
            d = json.loads(jsontext.decode())
            if "compresslevel" in d:
                return self.array_open(name)
            else:
                return self.dataset_open(name)

    def delete(self, name: str):
        jsontext = self._storage.get_or_none(os.path.join(name, "info.json"))
        if jsontext is None:
            if self._storage.get_or_none(name) is not None:
                self.blob_delete(name)
            else:
                raise Exception(f"Path {name} is invalid for deletion")
        else:
            d = json.loads(jsontext.decode())
            if "compresslevel" in d:
                self.array_delete(name)
            else:
                self.dataset_delete(name)

    def array_create(
        self,
        name: str,
        shape: Iterable[int],
        chunk: Iterable[int],
        dtype: Union[str, np.dtype],
        compress: str = "default",
        compresslevel: float = 0.5,
        dsplit: int = None,
    ) -> Array:

        shape = tuple(shape)
        chunk = tuple(chunk)

        overwrite = False
        props = Props()
        props.shape = shape
        props.chunk = chunk
        props.dtype = np.dtype(dtype).str
        props.compress = compress
        props.compresslevel = compresslevel
        if dsplit is not None:
            assert isinstance(dsplit, int)
            props.darray = "darray"
            darray_path = os.path.join(name, props.darray)
            darray_shape = shape[:dsplit] + (len(shape) - dsplit,)
            darray_chunk = chunk[:dsplit] + (len(shape) - dsplit,)
            arr = self.array_create(darray_path, darray_shape, darray_chunk, "int32")
            slices = tuple(map(lambda s: slice(0, s), shape[:dsplit]))
            arr[slices] = shape[dsplit:]

        assert len(shape) == len(chunk)

        info_path = os.path.join(name, "info.json")
        chunk_path = os.path.join(name, "chunks")

        if overwrite or not self._storage.exists(info_path):
            self._storage.put(info_path, bytes(json.dumps(props.__dict__), "utf-8"))

        if overwrite and self._storage.exists(chunk_path):
            self._storage.delete(chunk_path)

        return Array(name, self._storage)

    def array_open(self, name: str) -> Array:
        return Array(name, self._storage)

    def array_delete(self, name: str):
        self._storage.delete(name)

    def dataset_create(self, name: str, components: Dict[str, str]) -> Dataset:
        overwrite = False
        props = DatasetProps()
        props.paths = {}
        for key, comp in components.items():
            if isinstance(comp, Array):
                props.paths[key] = comp._path
            elif isinstance(comp, str):
                props.paths[key] = comp
            else:
                raise Exception(
                    "Input to the dataset is unknown: {}:{}".format(k, value)
                )

        dataset_path = os.path.join(name, "info.json")

        if overwrite or not self._storage.exists(dataset_path):
            self._storage.put(dataset_path, bytes(json.dumps(props.__dict__), "utf-8"))

        return Dataset(name, self._storage)

    def dataset_open(self, name: str) -> Dataset:
        return Dataset(name, self._storage)

    def dataset_delete(self, name: str):
        return self._storage.delete(name)

    def blob_get(self, name: str, none_on_errors: bool = False) -> bytes:
        if none_on_errors:
            return self._storage.get_or_none(name)
        else:
            return self._storage.get(name)

    def blob_set(self, name: str, content: bytes):
        self._storage.put(name, content)

    def blob_delete(self, name: str) -> bytes:
        self._storage.delete(name)

    def pytorch_model_get(self, name: str, none_on_errors: bool = False) -> dict:
        blob = self.blob_get(name, none_on_errors)

        if blob is None and not none_on_errors:
            raise Error("Error loading pytorch model")
        elif blob is None:
            return None
        else:
            f = io.BytesIO(blob)
            return torch.load(f)

    def pytorch_model_set(self, name: str, model: dict):
        f = io.BytesIO()
        torch.save(model, f)
        f.seek(0)
        self.blob_set(name, f.read())

    def pytorch_model_delete(self, name: str):
        self.blob_delete(name)



# BSD 3-Clause License

# Copyright (c) 2017, Ignacio Tartavull, William Silversmith, and later authors.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#  contributors may be used to endorse or promote products derived from
#  this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function
from six.moves import range, reduce

import json
import os
import re
import sys
import math
import operator
import time
import random
import string
from itertools import product
import posixpath

import numpy as np
from hub.exceptions import OutOfBoundsError, AlignmentError


def map2(fn, a, b):
    assert len(a) == len(b), "Vector lengths do not match: {} (len {}), {} (len {})".format(
        a[:3], len(a), b[:3], len(b))

    result = np.empty(len(a))

    for i in range(len(result)):
        result[i] = fn(a[i], b[i])

    if isinstance(a, Vec) or isinstance(b, Vec):
        return Vec(*result)

    return result


def max2(a, b):
    return map2(max, a, b).astype(a.dtype)


def min2(a, b):
    return map2(min, a, b).astype(a.dtype)


def clamp(val, low, high):
    return min(max(val, low), high)


def generate_chunks(img, offset, chunk_size):
    shape = Vec(*img.shape)
    offset = Vec(*offset)

    bounds = Bbox(offset, shape + offset)

    alignment_check = bounds.round_to_chunk_size(
        chunk_size, Vec.zeros(*chunk_size))

    if not np.all(alignment_check.minpt == bounds.minpt):
        raise AlignmentError("""
        Only chunk aligned writes are supported by this function. 
  
        Got:             {}
        Volume Offset:   {} 
        Nearest Aligned: {}
      """.format(
            bounds, Vec.zeros(*chunk_size), alignment_check)
        )

    bounds = Bbox.clamp(bounds, bounds)

    img_offset = bounds.minpt - offset
    img_end = Vec.clamp(bounds.size3() + img_offset, Vec.zeros(shape), shape)

    for startpt in product(*xranges):
        startpt = startpt.clone()
        endpt = min2(startpt + chunk_size, shape)
        spt = (startpt + bounds.minpt).astype(int)
        ept = (endpt + bounds.minpt).astype(int)
        yield (startpt, endpt, spt, ept)


def shade(dest_img, dest_bbox, src_img, src_bbox):
    """
    Shade dest_img at coordinates dest_bbox using the
    image contained in src_img at coordinates src_bbox.

    The buffer will only be painted in the overlapping
    region of the content.

    Returns: void
    """
    if not Bbox.intersects(dest_bbox, src_bbox):
        return

    spt = max2(src_bbox.minpt, dest_bbox.minpt)
    ept = min2(src_bbox.maxpt, dest_bbox.maxpt)

    dbox = Bbox(spt, ept) - dest_bbox.minpt
    sbox = Bbox(spt, ept) - src_bbox.minpt
 
    try:
        dest_img[dbox.to_slices()] = src_img[sbox.to_slices()]
    except Exception as e :
        print('error found {}'.format(str(e)))
        #print(dbox.to_slices(), sbox.to_slices())
        exit()


def chunknames(bbox, volume_bbox, key, chunk_size, protocol=None):

    path = posixpath if protocol != 'file' else os.path

    xranges = []
    for minim, maxim, step_size in zip(bbox.minpt, bbox.maxpt, chunk_size):
        a = range(minim, maxim, step_size)
        xranges.append(a)

    for index in product(*xranges):
        highpt = min2(Vec(*index) + chunk_size,
                      Bbox([0 for i in volume_bbox], volume_bbox).maxpt)
        filename = ["{}-{}".format(x, max_x)
                    for x, max_x in zip(index, highpt)]
        filename = "_".join(filename)
        yield path.join(key, 'chunks', filename)


def check_bounds(val, low, high):
    if val > high or val < low:
        raise OutOfBoundsError(
            'Value {} cannot be outside of inclusive range {} to {}'.format(val, low, high))
    return val


if sys.version_info < (3,):
    integer_types = (int, long, np.integer)
else:
    integer_types = (int, np.integer)

floating_types = (float, np.floating)

# Formula produces machine epsilon regardless of platform architecture
MACHINE_EPSILON = (7. / 3) - (4. / 3) - 1


def floating(lst):
    return any((isinstance(x, float) for x in lst))


class Vec(np.ndarray):
    def __new__(cls, *args, **kwargs):
        dtype = kwargs['dtype'] if 'dtype' in kwargs else int
        return super(Vec, cls).__new__(cls, shape=(len(args),), buffer=np.array(args).astype(dtype), dtype=dtype)

    @classmethod
    def clamp(cls, val, minvec, maxvec):
        return Vec(*min2(max2(val, minvec), maxvec))

    @classmethod
    def zeros(cls, shape):
        if isinstance(shape, int):
            shape = range(shape)
        return Vec(*[0 for i in shape])

    def clone(self):
        return Vec(*self[:], dtype=self.dtype)

    def null(self):
        return self.length() <= 10 * np.finfo(np.float32).eps

    def dot(self, vec):
        return sum(self * vec)

    def length2(self):
        return self.dot(self)

    def length(self):
        return math.sqrt(self.dot(self))

    def rectVolume(self):
        return reduce(operator.mul, self)

    def __hash__(self):
        return int(''.join(map(str, self)))

    def __repr__(self):
        values = u",".join(list(self.astype(str)))
        return u"Vec({}, dtype={})".format(values, self.dtype)


class Bbox(object):
    __slots__ = ['minpt', 'maxpt', '_dtype']

    """Represents a three dimensional cuboid in space."""

    def __init__(self, a, b, dtype=None):
        if dtype is None:
            if floating(a) or floating(b):
                dtype = np.float32
            else:
                dtype = np.int32

        self.minpt = Vec(*[min(ai, bi) for ai, bi in zip(a, b)], dtype=dtype)
        self.maxpt = Vec(*[max(ai, bi) for ai, bi in zip(a, b)], dtype=dtype)

        self._dtype = np.dtype(dtype)

    @classmethod
    def deserialize(cls, bbx_data):
        bbx_data = json.loads(bbx_data)
        return Bbox.from_dict(bbx_data)

    def serialize(self):
        return json.dumps(self.to_dict())

    @property
    def ndim(self):
        return len(self.minpt)

    @property
    def dtype(self):
        return self._dtype

    @classmethod
    def intersection(cls, bbx1, bbx2):
        result = Bbox([0] * bbx1.ndim, [0] * bbx2.ndim)

        if not Bbox.intersects(bbx1, bbx2):
            return result

        for i in range(result.ndim):
            result.minpt[i] = max(bbx1.minpt[i], bbx2.minpt[i])
            result.maxpt[i] = min(bbx1.maxpt[i], bbx2.maxpt[i])

        return result

    @classmethod
    def intersects(cls, bbx1, bbx2):
        return np.all(bbx1.minpt < bbx2.maxpt) and np.all(bbx1.maxpt > bbx2.minpt)

    @classmethod
    def near_edge(cls, bbx1, bbx2, distance=0):
        return (
            np.any(np.abs(bbx1.minpt - bbx2.minpt) <= distance)
            or np.any(np.abs(bbx1.maxpt - bbx2.maxpt) <= distance)
        )

    @classmethod
    def create(cls, obj, context=None, bounded=False):
        typ = type(obj)
        if typ is Bbox:
            obj = obj
        elif typ in (list, tuple):
            obj = Bbox.from_slices(obj, context, bounded)
        elif typ is Vec:
            obj = Bbox.from_vec(obj)
        elif typ is str:
            obj = Bbox.from_filename(obj)
        elif typ is dict:
            obj = Bbox.from_dict(obj)
        else:
            raise NotImplementedError(
                "{} is not a Bbox convertible type.".format(typ))

        if context and bounded:
            if not context.contains_bbox(obj):
                raise OutOfBoundsError(
                    "{} did not fully contain the specified bounding box {}.".format(
                        context, obj
                    ))

        return obj

    @classmethod
    def from_delta(cls, minpt, plus):
        return Bbox(minpt, Vec(*minpt) + plus)

    @classmethod
    def from_dict(cls, data):
        dtype = data['dtype'] if 'dtype' in data else np.float32
        return Bbox(data['minpt'], data['maxpt'], dtype=dtype)

    @classmethod
    def from_vec(cls, vec, dtype=int):
        return Bbox((0, 0, 0), vec, dtype=dtype)

    @classmethod
    def from_filename(cls, filename, dtype=int):
        match = os.path.basename(filename).replace(
            r'(?:\.gz)?$', '').split('_')
        match = [x.split('-') for x in match]
        mins = [int(x[0]) for x in match]
        maxs = [int(x[1]) for x in match]

        return Bbox(mins, maxs, dtype=dtype)

    @classmethod
    def from_slices(cls, slices, context=None, bounded=False):
        if context:
            slices = context.reify_slices(slices, bounded=bounded)

        return Bbox(
            [slc.start for slc in slices],
            [slc.stop for slc in slices]
        )

    @classmethod
    def from_list(cls, lst):
        """
        from_list(cls, lst)
        the lst length should be 6
        the first three values are the start, and the last 3 values are the stop 
        """
        assert len(lst) == 6
        return Bbox(lst[:3], lst[3:6])

    def to_filename(self):
        return '_'.join(
            (str(self.minpt[i]) + '-' + str(self.maxpt[i])
             for i in range(self.ndim))
        )

    def to_slices(self):
        return tuple([
            slice(int(self.minpt[i]), int(self.maxpt[i])) for i in range(self.ndim)
        ])

    def to_list(self):
        return list(self.minpt) + list(self.maxpt)

    def to_dict(self):
        return {
            'minpt': self.minpt.tolist(),
            'maxpt': self.maxpt.tolist(),
            'dtype': np.dtype(self.dtype).name,
        }

    def to_shape(self):
        return list(self.maxpt - self.minpt)

    def reify_slices(self, slices, bounded=True):
        """
        Convert free attributes of a slice object 
        (e.g. None (arr[:]) or Ellipsis (arr[..., 0]))
        into bound variables in the context of this
        bounding box.

        That is, for a ':' slice, slice.start will be set
        to the value of the respective minpt index of 
        this bounding box while slice.stop will be set 
        to the value of the respective maxpt index.

        Example:
          bbx = Bbox( (-1,-2,-3), (1,2,3) )
          bbx.reify_slices( (np._s[:],) )

          >>> [ slice(-1,1,1), slice(-2,2,1), slice(-3,3,1) ]

        Returns: [ slice, ... ]
        """
        if isinstance(slices, integer_types) or isinstance(slices, floating_types):
            slices = [slice(int(slices), int(slices)+1, 1)]
        elif type(slices) == slice:
            slices = [slices]
        elif type(slices) == Bbox:
            slices = slices.to_slices()
        elif slices == Ellipsis:
            slices = []

        slices = list(slices)

        for index, slc in enumerate(slices):
            if slc == Ellipsis:
                fill = self.ndim - len(slices) + 1
                slices = slices[:index] + \
                    (fill * [slice(None, None, None)]) + slices[index+1:]
                break

        while len(slices) < self.ndim:
            slices.append(slice(None, None, None))

        # First three slices are x,y,z, last is channel.
        # Handle only x,y,z here, channel seperately
        for index, slc in enumerate(slices):
            if isinstance(slc, integer_types) or isinstance(slc, floating_types):
                slices[index] = slice(int(slc), int(slc)+1, 1)
            elif slc == Ellipsis:
                raise ValueError(
                    "More than one Ellipsis operator used at once.")
            else:
                start = self.minpt[index] if slc.start is None else slc.start
                end = self.maxpt[index] if slc.stop is None else slc.stop
                step = 1 if slc.step is None else slc.step

                if step < 0:
                    raise ValueError(
                        'Negative step sizes are not supported. Got: {}'.format(step))

                # note: when unbounded, negative indicies do not refer to
                # the end of the volume as they can describe, e.g. a 1px
                # border on the edge of the beginning of the dataset as in
                # marching cubes.
                if bounded:
                    # if start < 0: # this is support for negative indicies
                    # start = self.maxpt[index] + start
                    check_bounds(start, self.minpt[index], self.maxpt[index])
                    # if end < 0: # this is support for negative indicies
                    #   end = self.maxpt[index] + end
                    check_bounds(end, self.minpt[index], self.maxpt[index])

                slices[index] = slice(start, end, step)

        return slices

    @classmethod
    def expand(cls, *args):
        result = args[0].clone()
        for bbx in args:
            result.minpt = min2(result.minpt, bbx.minpt)
            result.maxpt = max2(result.maxpt, bbx.maxpt)
        return result

    @classmethod
    def clamp(cls, bbx0, bbx1):
        result = bbx0.clone()
        result.minpt = Vec.clamp(bbx0.minpt, bbx1.minpt, bbx1.maxpt)
        result.maxpt = Vec.clamp(bbx0.maxpt, bbx1.minpt, bbx1.maxpt)
        return result

    def size(self):
        return Vec(*(self.maxpt - self.minpt), dtype=self.dtype)

    def size3(self):
        return Vec(*(self.maxpt[:3] - self.minpt[:3]), dtype=self.dtype)

    def subvoxel(self):
        """
        Previously, we used bbox.volume() < 1 for testing
        if a bounding box was larger than one voxel. However, 
        if two out of three size dimensions are negative, the 
        product will be positive. Therefore, we first test that 
        the maxpt is to the right of the minpt before computing 
        whether conjunctioned with volume() < 1.

        Returns: boolean
        """
        return (not self.valid()) or self.volume() < 1

    def empty(self):
        """
        Previously, we used bbox.volume() <= 0 for testing
        if a bounding box was empty. However, if two out of 
        three size dimensions are negative, the product will 
        be positive. Therefore, we first test that the maxpt 
        is to the right of the minpt before computing whether 
        the bbox is empty and account for 20x machine epsilon 
        of floating point error.

        Returns: boolean
        """
        return (not self.valid()) or (self.volume() < (20 * MACHINE_EPSILON))

    def valid(self):
        return np.all(self.minpt <= self.maxpt)

    def volume(self):
        if np.issubdtype(self.dtype, np.integer):
            return self.size3().astype(np.int64).rectVolume()
        else:
            return self.size3().astype(np.float64).rectVolume()

    def center(self):
        return (self.minpt + self.maxpt) / 2.0

    def grow(self, amt):
        assert amt >= 0
        self.minpt -= amt
        self.maxpt += amt
        return self

    def shrink(self, amt):
        assert amt >= 0
        self.minpt += amt
        self.maxpt -= amt

        if not self.valid():
            raise ValueError("Cannot shrink bbox below zero volume.")

        return self

    def expand_to_chunk_size(self, chunk_size, offset=Vec(0, 0, 0, dtype=int)):
        """
        Align a potentially non-axis aligned bbox to the grid by growing it
        to the nearest grid lines.

        Required:
          chunk_size: arraylike (x,y,z), the size of chunks in the 
                        dataset e.g. (64,64,64)
        Optional:
          offset: arraylike (x,y,z), the starting coordinate of the dataset
        """
        chunk_size = np.array(chunk_size, dtype=np.float32)
        result = self.clone()
        result = result - offset
        result.minpt = np.floor(result.minpt / chunk_size) * chunk_size
        result.maxpt = np.ceil(result.maxpt / chunk_size) * chunk_size
        return (result + offset).astype(self.dtype)

    def shrink_to_chunk_size(self, chunk_size, offset=Vec(0, 0, 0, dtype=int)):
        """
        Align a potentially non-axis aligned bbox to the grid by shrinking it
        to the nearest grid lines.

        Required:
          chunk_size: arraylike (x,y,z), the size of chunks in the 
                        dataset e.g. (64,64,64)
        Optional:
          offset: arraylike (x,y,z), the starting coordinate of the dataset
        """
        chunk_size = np.array(chunk_size, dtype=np.float32)
        result = self.clone()
        result = result - offset
        result.minpt = np.ceil(result.minpt / chunk_size) * chunk_size
        result.maxpt = np.floor(result.maxpt / chunk_size) * chunk_size

        # If we are inside a single chunk, the ends
        # can invert, which tells us we should collapse
        # to a single point.
        if np.any(result.minpt > result.maxpt):
            result.maxpt = result.minpt.clone()

        return (result + offset).astype(self.dtype)

    def round_to_chunk_size(self, chunk_size, offset=Vec(0, 0, 0, dtype=int)):
        """
        Align a potentially non-axis aligned bbox to the grid by rounding it
        to the nearest grid lines.

        Required:
          chunk_size: arraylike (x,y,z), the size of chunks in the 
                        dataset e.g. (64,64,64)
        Optional:
          offset: arraylike (x,y,z), the starting coordinate of the dataset
        """
        chunk_size = np.array(chunk_size, dtype=np.float32)
        result = self.clone()
        result = result - offset
        result.minpt = np.round(result.minpt / chunk_size) * chunk_size
        result.maxpt = np.round(result.maxpt / chunk_size) * chunk_size
        return (result + offset).astype(self.dtype)

    def contains(self, point):
        """
        Tests if a point on or within a bounding box.

        Returns: boolean
        """
        return np.all(point >= self.minpt) and np.all(point <= self.maxpt)

    def contains_bbox(self, bbox):
        return self.contains(bbox.minpt) and self.contains(bbox.maxpt)

    def clone(self):
        return Bbox(self.minpt, self.maxpt, dtype=self.dtype)

    def astype(self, typ):
        tmp = self.clone()
        tmp.minpt = tmp.minpt.astype(typ)
        tmp.maxpt = tmp.maxpt.astype(typ)
        tmp._dtype = tmp.minpt.dtype
        return tmp

    def transpose(self):
        return Bbox(self.minpt[::-1], self.maxpt[::-1])

    # note that operand can be a vector
    # or a scalar thanks to numpy
    def __sub__(self, operand):
        tmp = self.clone()

        if isinstance(operand, Bbox):
            tmp.minpt -= operand.minpt
            tmp.maxpt -= operand.maxpt
        else:
            tmp.minpt -= operand
            tmp.maxpt -= operand

        return tmp

    def __iadd__(self, operand):
        if isinstance(operand, Bbox):
            self.minpt += operand.minpt
            self.maxpt += operand.maxpt
        else:
            self.minpt += operand
            self.maxpt += operand

        return self

    def __add__(self, operand):
        tmp = self.clone()
        return tmp.__iadd__(operand)

    def __imul__(self, operand):
        self.minpt *= operand
        self.maxpt *= operand
        return self

    def __mul__(self, operand):
        tmp = self.clone()
        tmp.minpt *= operand
        tmp.maxpt *= operand
        return tmp.astype(tmp.minpt.dtype)

    def __idiv__(self, operand):
        if (
            isinstance(operand, float)
            or self.dtype in (float, np.float32, np.float64)
            or (hasattr(operand, 'dtype') and operand.dtype in (float, np.float32, np.float64))
        ):
            return self.__itruediv__(operand)
        else:
            return self.__ifloordiv__(operand)

    def __div__(self, operand):
        if (
            isinstance(operand, float)
            or self.dtype in (float, np.float32, np.float64)
            or (hasattr(operand, 'dtype') and operand.dtype in (float, np.float32, np.float64))
        ):

            return self.__truediv__(operand)
        else:
            return self.__floordiv__(operand)

    def __ifloordiv__(self, operand):
        self.minpt //= operand
        self.maxpt //= operand
        return self

    def __floordiv__(self, operand):
        tmp = self.astype(float)
        tmp.minpt //= operand
        tmp.maxpt //= operand
        return tmp.astype(int)

    def __itruediv__(self, operand):
        self.minpt /= operand
        self.maxpt /= operand
        return self

    def __truediv__(self, operand):
        tmp = self.clone()

        if isinstance(operand, int):
            operand = float(operand)

        tmp.minpt = Vec(*(tmp.minpt.astype(float) / operand), dtype=float)
        tmp.maxpt = Vec(*(tmp.maxpt.astype(float) / operand), dtype=float)
        return tmp.astype(tmp.minpt.dtype)

    def __ne__(self, other):
        return not (self == other)

    def __eq__(self, other):
        return np.array_equal(self.minpt, other.minpt) and np.array_equal(self.maxpt, other.maxpt)

    def __hash__(self):
        return int(''.join(map(str, map(int, self.to_list()))))

    def __repr__(self):
        return "Bbox({},{}, dtype={})".format(list(self.minpt), list(self.maxpt), self.dtype)



import hub


def test_dataset():
    datahub = hub.fs("./data/cache").connect()
    x = datahub.array(
        name="test/example:input", shape=(100, 25, 25), chunk=(20, 5, 5), dtype="uint8"
    )
    y = datahub.array(
        name="test/example:label", shape=(100, 4), chunk=(20, 2), dtype="uint8"
    )

    ds = datahub.dataset(
        components={"input": x, "label": y}, name="test/dataset:train3"
    )
    assert ds[0]["input"].shape == (25, 25)
    assert ds["input"].shape[0] == 100  # return single array
    assert ds["label", 0].mean() == 0  # equivalent ds['train'][0]


def test_load_dataset():
    datahub = hub.fs("./data/cache").connect()
    ds = datahub.open(name="test/dataset:train3")  # return the dataset object
    assert list(map(lambda x: x[0], ds.items())) == ["input", "label"]
    assert ds["label", 0].mean() == 0



from typing import *
import sys, os, json, uuid, traceback, random, time
import numpy as np

# from pathos.threading import ThreadPool

from multiprocessing.pool import ThreadPool

from .storage import Base as Storage
from . import codec
from .bbox import Bbox, chunknames, shade, Vec, generate_chunks
from .exceptions import (
    IncompatibleBroadcasting,
    IncompatibleTypes,
    IncompatibleShapes,
    NotFound,
)

_hub_thread_pool = None


class Props:
    shape: Tuple[int, ...] = None
    chunk_shape: Tuple[int, ...] = None
    dtype: str = None
    compress: str = None
    compresslevel: float = 0.5
    # dsplit: Optional[Union[int, List[int]]] = None
    darray: str = None

    @property
    def chunk(self) -> Tuple[int, ...]:
        return self.chunk_shape

    @chunk.setter
    def chunk(self, value: Tuple[int, ...]):
        self.chunk_shape = value

    def __init__(self, dict: dict = None):
        if dict is not None:
            self.__dict__ = dict


class Array:
    def __init__(self, path: str, storage: Storage, threaded=False):
        self._path = path
        self._storage = storage
        self._props = Props(json.loads(storage.get(os.path.join(path, "info.json"))))
        self._codec = codec.from_name(self.compress, self.compresslevel)
        self._dcodec = codec.Default()
        global _hub_thread_pool
        if _hub_thread_pool is None and threaded:
            print("Thread Pool Created")
            _hub_thread_pool = ThreadPool(32)
        self._map = _hub_thread_pool.map if threaded else map

        self._darray = None
        if self._props.darray:
            self._darray = Array(os.path.join(path, self._props.darray), storage)
        # assert isinstance(self._props.dsplit, int)

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._props.shape

    @property
    def darray(self) -> "Array":
        return self._darray

    # @property
    # def dynamic_shape(self, slices: Union[Slice[int, ...], Tuple[int, ...], List[int, ...]]):
    #     pass

    @property
    def chunk(self) -> Tuple[int, ...]:
        return self._props.chunk

    @property
    def dtype(self) -> str:
        return self._props.dtype

    @property
    def compress(self) -> str:
        return self._props.compress

    @property
    def compresslevel(self) -> float:
        return self._props.compresslevel

    # @property
    # def _dshape_path(self):
    #     return os.path.join(self._path, 'dshape.json')

    # def _get_dshape(self) -> np.ndarray:
    #     data = self._storage.get(self._dshape_path)
    #     return self._dcodec.decode(data)

    # def _set_dshape(self, arr: np.ndarray):
    #     data = self._dcodec.encode(arr)
    #     self._storage.put(self._dshape_path, data)

    # def get_shape(self, slices: Iterable[slice]):
    #     slices = tuple(slices)

    # # Iterable slices
    # def set_shape(self, slices: Iterable[slice], shape: Iterable[int]):
    #     slices = tuple(slices)
    #     shape = tuple(shape)
    #     assert len(slices) == self._props.dsplit
    #     assert len(shape) == len(self._props.shape) - self._props.dsplit

    #     arr = self._get_dshape()
    #     arr[slices] = shape
    #     self._set_dshape(arr)

    def __getitem__(self, slices: Tuple[slice]):
        cloudpaths, requested_bbox = self._generate_cloudpaths(slices)
        tensor = self._download(cloudpaths, requested_bbox)
        tensor = self._squeeze(slices, tensor)
        return tensor

    def __setitem__(self, slices: Tuple[slice], content: np.ndarray):
        cloudpaths, requested_bbox = self._generate_cloudpaths(slices)
        self._upload(cloudpaths, requested_bbox, content)

    def _generate_cloudpaths(self, slices):
        # Slices -> Bbox
        if isinstance(slices, int):
            slices = (slices,)
        elif isinstance(slices, slice):
            slices = (slices,)
        slices = tuple(slices)

        _shape = list(self.shape)
        if self._darray is not None:
            s = len(self._darray.shape) - 1
            arr = self._darray[slices[:s]]
            res = np.amax(arr, axis=tuple(range(0, len(arr.shape) - 1)))
            assert len(res.shape) == 1
            assert len(_shape[s:]) == res.shape[0]
            _shape[s:] = res
            _shape = tuple(_shape)

        slices = Bbox(Vec.zeros(_shape), _shape).reify_slices(slices, bounded=True)
        requested_bbox = Bbox.from_slices(slices)

        # Make sure chunks fit
        full_bbox = requested_bbox.expand_to_chunk_size(
            self.chunk, offset=Vec.zeros(self.shape)
        )

        # Clamb the border
        full_bbox = Bbox.clamp(full_bbox, Bbox(Vec.zeros(self.shape), self.shape))

        # Generate chunknames
        cloudpaths = list(
            chunknames(
                full_bbox,
                self.shape,
                self._path,
                self.chunk,
                protocol="none",  # self.protocol
            )
        )

        return cloudpaths, requested_bbox

    def _download_chunk(self, cloudpath):

        chunk = self._storage.get_or_none(cloudpath)
        if chunk:
            chunk = self._codec.decode(chunk)
        else:
            chunk = np.zeros(shape=self.chunk, dtype=self.dtype)

        bbox = Bbox.from_filename(cloudpath)
        return chunk, bbox

    def _download(self, cloudpaths, requested_bbox):
        # Download chunks
        chunks_bboxs = list(self._map(self._download_chunk, cloudpaths))

        # Combine Chunks
        renderbuffer = np.zeros(shape=requested_bbox.to_shape(), dtype=self.dtype)

        def process(chunk_bbox):
            chunk, bbox = chunk_bbox
            shade(renderbuffer, requested_bbox, chunk, bbox)

        list(self._map(process, chunks_bboxs))

        return renderbuffer

    def _squeeze(self, slices, tensor):
        squeeze_dims = []

        if isinstance(slices, list) and len(slices) == 1:
            slices = slices[0]

        if not isinstance(slices, list) and not isinstance(slices, tuple):
            slices = [slices]

        for dim in range(len(slices)):
            if isinstance(slices[dim], int):
                squeeze_dims.append(dim)

        if len(squeeze_dims) >= 1:
            tensor = tensor.squeeze(axis=(*squeeze_dims,))

        if len(tensor.shape) == 0:
            tensor = tensor.item()

        return tensor

    def _upload_chunk(self, cloudpath_chunk):
        cloudpath, chunk = cloudpath_chunk
        chunk = self._codec.encode(chunk)
        chunk = self._storage.put(cloudpath, chunk)

    def _chunkify(self, cloudpaths, requested_bbox, item):
        chunks = []
        for path in cloudpaths:
            cloudchunk = Bbox.from_filename(path)
            intersection = Bbox.intersection(cloudchunk, requested_bbox)
            chunk_slices = (intersection - cloudchunk.minpt).to_slices()
            item_slices = (intersection - requested_bbox.minpt).to_slices()

            chunk = None
            if np.any(np.array(intersection.to_shape()) != np.array(self.chunk)):
                chunk, _ = self._download_chunk(path)
            else:
                chunk = np.zeros(shape=self.chunk, dtype=self.dtype)

            chunk.setflags(write=1)
            chunk[chunk_slices] = item[item_slices]
            chunks.append(chunk)

        return zip(cloudpaths, chunks)

    def _upload(self, cloudpaths, requested_bbox, item):
        try:
            item = np.broadcast_to(item, requested_bbox.to_shape())
        except ValueError as err:
            raise IncompatibleBroadcasting(err)

        try:
            item = item.astype(self.dtype)
        except Exception as err:
            raise IncompatibleTypes(err)

        cloudpaths_chunks = self._chunkify(cloudpaths, requested_bbox, item)
        list(self._map(self._upload_chunk, list(cloudpaths_chunks)))




from .creds import Base as Creds
from .array import Array
from .bucket import Bucket
from .dataset import Dataset

s3 = Creds._s3
gs = Creds._gs
fs = Creds._fs

# @click.group()
# @click.option('-v', '--verbose', count=True, help='Devel debugging')
# def cli(verbose):
#     configure_logger(verbose)

# def add_commands(cli):
#     cli.add_command(configure)

# add_commands(cli)
